/**
 * BloomCounter is a counter that is assisted with a bloom filter
 * for speed. The first insertion just puts it in the bloom. The
 * second adds twice to the map.
 *
 * HashingBloomCounter also computes the hash of the block; it 
 * hash special logic for blocks of all NULLs.
 * 
 * Both of these objects have sha1() and md5() covers. The sha1()
 * is more secure and implements the NIST standard, but md5() is
 * significantly faster.
 */

#ifndef _BLOOMCOUNTER_H
#define _BLOOMCOUNTER_H

#include "sha1_t.h"
#include <set>
#include <map>

class sha1set:public std::set<class sha1_t> {};

class BloomCounter : public NSRLBloom {
public:
 BloomCounter():total_adds(0){ }
    class sha1counter_t:public std::map<class sha1_t,uint64_t> { };
    sha1counter_t counter;
    uint64_t total_adds;
    void add_sha1(const sha1_t &sha1){
	total_adds ++;
	if(NSRLBloom::query(sha1.final())){
	    if(counter[sha1]==0) counter[sha1]=1;
	    counter[sha1]+=1;
	}
	else {
	    NSRLBloom::add(sha1.final());
	}
    }
};

class HashingBloomCounter : public BloomCounter {
 private:
    const EVP_MD *md;
    static bool iszero(const u_char *buf,size_t bufsize){
	for(u_int i=0;i<bufsize;i++){
	    if(buf[i]!=0) return false;
	}
	return true;
    }
 public:
    uint64_t zero_count;
    size_t zero_size;
 HashingBloomCounter():zero_count(0),zero_size(0){
    }
    void add_buf(const u_char *buf,size_t bufsize){
	if(zero_size==0 || bufsize==zero_size){
	    if(iszero(buf,bufsize)){
		zero_size = bufsize;
		zero_count ++;
		return;
	    }
	}
	const sha1_t ch = sha1gen_t::hash_buf(buf,bufsize);
	add_sha1(ch);
    }
};

#endif
