

/*
 * precarve: given a disk image and a bloom filter, remove all of the sectors in the disk image
 * that have hashes that match the bloom filter.
 *
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>


#include "config.h"
#include "bloom.h"
#include "aftimer.h"
#include "err.h"
#include "threaded_hash.h"

#ifdef HAVE_LIBAFFLIB
#include <afflib/afflib.h>
#include <afflib/afflib_i.h>		// need af_imagepagesize; easy way to get individual pages
#include <afflib/utils.h>

#include <fcntl.h>
#include <iostream>
#include <bitset>
#include <assert.h>

using namespace std;
u_int opt_run = 8;
int debug=0;

#define SECTOR_SIZE 512
#define MAX_SECTORS 1024*1024*16	// that's a lot of sectors

void usage()
{
    printf("usage: precarve [options] input.rawfilter.bf output.raw\n");
    printf("\n");
    printf("where input.raw  - input disk image\n");
    printf("      filter.bf  - filter to prune against (sectors in filter are not put in output.raw\n");
    printf("      output.raw - output file\n");
    printf("options:\n");
    printf("   -r nn   - require a run of nn duplicate blocks before a block is filtered out\n");
    printf("             (default is %d)\n",opt_run);
    printf("   -d      - debug\n");
    exit(1);
}

int main(int argc,char **argv)
{
    int ch;

    while ((ch = getopt(argc,argv,"h?r:d")) != -1){
	switch(ch){
	case 'd': debug++;break;
	case 'r': opt_run = atoi(optarg); break;	    
	case 'h':
	case '?':
	default:
	    usage();
	}
    }

    argc -= optind;
    argv += optind;

    /* Open the image file and get a list of the segments. */
    AFFILE *ain = af_open(argv[0],O_RDONLY,0777);
    if(!ain) err(1,"%s",argv[0]);
    class aff::seglist segs(ain);

    /* Open the bloom filter */
    NSRLBloom bloom;
    if(bloom.open(argv[1],O_RDONLY)) err(1,"%s",argv[2]);

    /* Open the output file */
    int out = open(argv[2],O_WRONLY|O_CREAT|O_EXCL|O_BINARY,0777);
    if(out<0) err(1,"%s",argv[2]);

    /* Now process each segment */
    unsigned char *data = (unsigned char *)calloc(ain->image_pagesize,1);
    threaded_hash sha1(EVP_get_digestbyname("sha1"),0); // don't be multi-threaded

    /* Get a hash of the null segment */
    threaded_hash sha1_null(EVP_get_digestbyname("sha1"),0);
    sha1_null.update(data,SECTOR_SIZE);

    for(aff::seglist::const_iterator it = segs.begin();
	it!=segs.end();
	it++){
	int64_t pagenumber = (*it).pagenumber();
	if(pagenumber>=0){
	    cout << "processing segment " << (*it).name << "...";
	    size_t data_len = ain->image_pagesize;
	    if(af_get_page(ain,pagenumber,data,&data_len)){
		err(1,"af_get_page(file=%s,page=%"PRId64")", af_filename(ain),pagenumber);
	    }

	    /* Go through the page and for each sector determine if it is in the target set */

	    const u_int sectors = data_len/SECTOR_SIZE;
	    assert(sectors < MAX_SECTORS);
	    class std::bitset<MAX_SECTORS> keep,okeep;
	    u_int i = 0;
	    for(i=0;i<sectors;i++){
		const u_char *buf = data+i*SECTOR_SIZE;
		
		sha1.clear();
		sha1.update(buf,SECTOR_SIZE);
		
		keep[i] = (sha1.final()==sha1_null.final()) || (bloom.query(sha1.final())==false);
		okeep[i] = keep[i];	// remember result
	    }
		    
	    /* Now keep all of the sectors for which the run length of not-to-keep is less than
	     * opt_run. Do this by looking for keep-to-not-keep transitions.
	     */
	    for(i=0;i<sectors;){
		bool prev_keep = (i==0 || keep[i-1]);
		if(prev_keep && keep[i]==0){
		    /* See if this not-keep run length is less than opt_run */
		    u_int j;
		    for(j=i;j<sectors && keep[j]==0;j++){
		    }
		    if(j-i<=opt_run){
			/* The run of sectors not to keep is equal to or
			 * less than opt_run, so keep them. */
			for(u_int k=i;k<j && k<sectors;k++){
			    keep[k] = 1;
			}
		    }
		    i = j;		// we have processed ot the end of this run
		}
		else {
		    i++;
		}
	    }

	    /* Now copy over the sectors we are keeping */
	    int kept=0;
	    for(i=0;i<sectors;i++){
		if(keep[i]){
		    const u_char *buf = data+i*SECTOR_SIZE;
		    if(write(out,buf,SECTOR_SIZE)!=SECTOR_SIZE) err(1,"write");
		    kept++;
		}
	    }
	    cout << " (" << kept << " out of " << sectors << " kept)\n";
	    if(debug){
		for(i=0;i<sectors;i++){
		    if(okeep[i] || keep[i]){
			printf("%5d   %d    %d\n",i,okeep[i] ? 1 : 0,keep[i] ? 1 : 0);
		    }
		}
	    }
	}
    }
}	
#else
int main(int argc,char **argv)
{
    fprintf(stderr,"%s currently requires AFFLIB\n",argv[0]);
    exit(1);
}
#endif

