/**
 * self_same_calc.cpp:
 * 
 * read a set of hash codes and report the histogram results.
 * May be applied to NSRL 4K hash sets or to the disk images.
 */

#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include "config.h"
#include "bloom.h"

#ifdef HAVE_LIBAFFLIB

#define BLOCK_SIZE_DEFAULT 512
#define BLOOM_BITS_DEFAULT 32
#define REPORTING_THRESHOLD 2

unsigned int opt_reporting_threshold = REPORTING_THRESHOLD;

#include "threaded_hash.h"
#include "sha1_t.h"
#include <afflib/afflib.h>
#include <afflib/afflib_i.h>
#include <afflib/utils.h>

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>

using namespace std;
#include <vector>
#include <iostream>
#include <set>
#include <getopt.h>
#include <ctype.h>

#include "BloomCounter.h"

/*
 * Sample line:
 * INSERT INTO blockhash (filekey, hashval, hashtype, blocksize, block) VALUES ('9AF318B4749D308A6366A18DF0A7802D83E50EEF', '166092F87CAE2625B2A735A4039709778C54F085', 'SHA1B4096', 4096, 48);
 */


const char *progname = "sector_hash";
int opt_x = 0;
bool opt_dups = true;
const sha1_t *opt_match = 0;
const sha1_t *justhash = 0;
int opt_remove_constant = 0;			// remove constant sectors
uint64_t constants_removed = 0;

uint64_t hashcount=0;
uint64_t filecount=0;


void print_stats(const string &fname) 
{
    printf("%s filecount:%"PRId64"  hashcount:%"PRId64"\n",fname.c_str(),filecount,hashcount);
}

uint64_t freq=1000*1000;

bool is_constant(const u_char *buf,int bufsize)
{
    u_char first = buf[0];
    for(int i=1;i<bufsize;i++){
	if(buf[i]!=first) return false;
    }
    return true;
}

void process_nsrl_4k(class HashingBloomCounter &b,string fname)
{
    const char *match="VALUES";
    FILE *infile = 0;
    bool pipe = false;
    char buf[4096];

    cout << "process_nsrl_4k(" << fname << ")\n";
    if(strstr(fname.c_str(),".zip")){
	string cmd = "unzip -p "+fname;
	cout << cmd << "\n";
	infile = popen(cmd.c_str(),"r");
	if(!infile) err(1,"%s",cmd.c_str());
	pipe = true;
    }
    if(!infile){
	infile = fopen(fname.c_str(),"r");
	if(!infile) err(1,"%s",fname.c_str());
    }

    int lastnl=0;
    int lastvalues=0;
    
    std::set<class sha1_t> fileset;	// sha1s seen in the file
    while(fgets(buf,sizeof(buf),infile)){
	if(buf[lastnl]!='\n'){		// assume each line is the same length unless we learn otherwise
	    char *cc = strchr(buf,'\n');
	    if(!cc) break;		// no complete line?
	    lastnl = cc-buf;
	}
	buf[lastnl]=0;
	if(strcmp(buf+lastvalues,match)!=0){
	    const char *cc = strstr(buf,match);
	    if(!cc) continue;		// doesn't have a VALUES line
	    lastvalues = cc-buf;
	}
	const char *data = buf+lastvalues;

	const sha1_t *filesha1  = sha1_t::new_from_hex(data+9);
	const sha1_t *blocksha1 = sha1_t::new_from_hex(data+53);

	int blocknumber = atoi(data+115);
	if(blocknumber==1){
	    filecount++;
	    fileset.clear();
	}

	if(opt_dups){
	    if(fileset.find(blocksha1->final()) ==fileset.end()){
		fileset.insert(blocksha1->final());
		b.add_sha1(*blocksha1);
	    }
	}
	if(justhash && *justhash == *blocksha1){
	    puts(buf);
	}

	hashcount++;
	if(hashcount % freq == 0) print_stats(fname);

	delete filesha1;
	delete blocksha1;
    }
    if(pipe) pclose(infile);
    else fclose(infile);
    print_stats(fname);
}

int64_t matches = 0;
void print_report(class HashingBloomCounter &b)
{
    if(opt_match){
	printf("Number of matching: %"PRId64"\n",matches);
    }
    if(opt_remove_constant){
	printf("constants_removed: %"PRId64"\n",constants_removed);
    }

    printf("total in HashingBloomCounter: %8"PRIu64"\n",b.total_adds); // total # of times add_sha1 called
    printf("added_items:     %8"PRIu64"\n",b.added_items); // total number of times any bit set
    /* Don't bother printing aliased; it should be 0 with the
       HashingBloomCounter */

    /* Now go through all of the dups. Report the most popular SHAs */
    class sha1_tally:public std::multimap<uint64_t,sha1_t> {};
    sha1_tally tally;
    int64_t dups = 0;
    for(HashingBloomCounter::sha1counter_t::const_iterator i = b.counter.begin();
	i != b.counter.end(); i++){
	uint64_t count = (*i).second;
	
	dups += count;
	if(count>=opt_reporting_threshold){
	    tally.insert( pair<uint64_t,sha1_t> (count,(*i).first) );
	}
    }
    printf("Number of all NULL %zd-byte blocks: %9"PRIu64"\n",
	   b.zero_size,b.zero_count);
    printf("Number of block hashes that are duplicated: %9"PRIu64" (%f%%)\n",
	   dups,((double)dups)*100.0/b.total_adds);
    printf("Number of block hashes that are distinct:     %9"PRIu64" (%f%%)\n",
	   b.total_adds-dups,((double)b.total_adds-dups)*100.0/b.total_adds);
    printf("Bloom filter utilization: %f%%\n",b.utilization()*100.0);;
    uint64_t prev = 0;
    uint64_t count_on_line = 0;
    std::map<uint64_t,uint64_t> histogram;
    uint64_t count_max = 0;
    bool first = true;
    for(sha1_tally::const_iterator i = tally.begin();
	i != tally.end();
	i++){
	uint64_t count = (*i).first;
	const sha1_t &hash = (*i).second;
	
	if(count!=prev){
	    if(!first){
		// terminate the last line
		printf("[%"PRId64"]\n",count_on_line); 
		count_on_line = 0;
	    }
	    first = false;
	    printf("%"PRIu64": ",count);
	    prev = count;
	}
	if(count_on_line<2){
	    cout << hash.hexdigest() << " ";
	}
	if(count_on_line==2){
	    printf("... ");
	}
	count_on_line++;
	histogram[count] = count_on_line; // will get incremented
	count_max = count;
    }
    printf("\n");
    printf("Histogram:\n");
    printf("n-Tuple\tHash Count\n");
    for(uint64_t i=0;i<count_max+1;i++){
	if(histogram.find(i)!=histogram.end()){
	    printf("%"PRId64"\t%"PRId64"\t",i,histogram[i]);
	    for(u_int j=0;j<60 && j<histogram[i];j++) putchar('*');
	    putchar('\n');
	}
    }

    printf("\n");
    printf("hash count: %"PRIu64"\n",hashcount);
    printf("file count: %"PRIu64"\n",filecount);
}

void process_imagefile(const char *fname,class HashingBloomCounter &b,int sector_size)
{
    OpenSSL_add_all_digests();
    
    AFFILE *af = af_open(fname,O_RDONLY,0);
    if(!af) af_err(1,"af_open(%s)",fname);
    aff::seglist segs(af);
    u_char *buf = (unsigned char *)calloc(af->image_pagesize,1); 

    for(aff::seglist::const_iterator i=segs.begin();i!=segs.end();i++){
	size_t bytes = af->image_pagesize;
	int64_t pagenum = i->pagenumber();
	if(pagenum<0) continue; // not a data segment
	fprintf(stderr,"Processing Page %"PRId64"\r",pagenum);
	if(af_get_page(af,pagenum,buf,&bytes)!=0){
 	    fprintf(stderr,"  ** can't read page %"PRId64"\n",pagenum);
	    continue;
	}
	for(size_t o=0;o<bytes;o+=sector_size){
	    const unsigned char *sector_buf = buf+o;
	    if(af_is_badsector(af,sector_buf)){
		continue;
	    }
	    if(opt_remove_constant && is_constant(sector_buf,sector_size)){
		constants_removed++;
		continue;
	    }

	    sha1_t ch = sha1gen_t::hash_buf(sector_buf,sector_size);
	    b.add_sha1(ch);

	    if(opt_match && ((*opt_match) == ch)){
		cout << "match SHA-1: " << opt_match->hexdigest() << "\n";
		cout << "block SHA-1: " << ch.hexdigest() << "\n";
		cout << "\n";
		cout << "Hash match #" << matches << " page " << pagenum << " page offset " << o
		     << " (abs offset " << pagenum*af->image_pagesize+o << "):\n";
		
		matches += 1;
		const int width=16;
		for(ssize_t i=0;i<sector_size;i+=width){
		    for(ssize_t j=i;j<i+width;j++){
			unsigned char cc = sector_buf[j];
			if(j==i+width/2) putchar(' ');
			printf("%02x ",cc);
		    }
		    for(ssize_t j=i;j<i+width;j++){
			char cc = sector_buf[j];
			if(j==i+width/2) putchar(' ');
			putchar(isprint(cc) ? cc : '.');
		    }
		    putchar('\n');
		}
		putchar('\n');
		putchar('\n');
	    }
	}
    }
}


void usage()
{
    printf("Usage: %s [options] \n", progname);
    printf("Calculates the self-same metric\n");
    printf("     -r file/dir  read the NIST SQL 4K block insert file and report stats.\n");
    printf("                  automatically decompresses .zip files and reads dirs\n");
    printf("     -a filename.aff Reads bock hashes from a disk image\n");
    printf("     -b nnn       Sets the block hash size for the disk file. Default is %d\n", BLOCK_SIZE_DEFAULT);
    printf("     -B nn        Sets the number of bloom bits; default is %d\n",BLOOM_BITS_DEFAULT);
    printf("     -j <hash>    Just report stats for a particular hash\n");
    printf("     -h           Shows this usage.\n");
    printf("     -w <file>    write the bloom file out to the specified file\n");
    printf("     -t nnn       Set the reporting threshold; default is %d\n",REPORTING_THRESHOLD);
    printf("     -m <hexhash> Report matches on hexhash\n");
    printf("     -R           Remove constant sectors\n");
    printf("\n");
}



int main(int argc,char **argv)
{
    printf("Command Line: ");
    for(int i=0;i<argc;i++){
	printf("%s ",argv[i]);
    }
    printf("\n");

    class HashingBloomCounter b;
    vector<string> rfiles;
    const char *outfile = 0;
    const char *opt_imagefile = 0;
    int opt_sector_size = BLOCK_SIZE_DEFAULT;
    int opt_bloom_bits = BLOOM_BITS_DEFAULT;
    int ch;
    while((ch = getopt(argc, argv, "j:r:w:h?b:B:a:t:m:R")) != -1){
        switch(ch){
	case 'a': opt_imagefile = optarg; break;
	case 'b': opt_sector_size = atoi(optarg); break;
	case 'B': opt_bloom_bits = atoi(optarg); break;
	case 'j':
	    opt_dups = false;
	    justhash = sha1_t::new_from_hex(optarg);
	    break;
	case 'r': rfiles.push_back(optarg); break;
	case 't': opt_reporting_threshold = atoi(optarg);break;
	case 'w': outfile = optarg; break;
	case 'm': opt_match = sha1_t::new_from_hex(optarg); break;
	case 'R': opt_remove_constant = 1;break;
	default:
	case '?':
	case 'h':
	    usage();
	    exit(0);
        }
    }

    argc -= optind;  
    argv += optind;  

    if(b.create(0,160,opt_bloom_bits,4,"Standard hash with 2^32 elements")){
	err(1,"%s: Cannot create bloom file: ",outfile);
    }
    for(vector<string>::const_iterator i = rfiles.begin();i!=rfiles.end();i++){
	// If it is a directory, process all of the zip files in the directory
	string fname = *i;
	cout << "processing " << fname << "\n";
	struct stat st;
	if((stat(fname.c_str(),&st)==0) && S_ISDIR(st.st_mode)){
	    DIR *dirp = opendir(fname.c_str());
	    struct dirent *dp;
	    while ((dp = readdir(dirp)) != NULL){
		if(strstr(dp->d_name,"blockhash_") || strstr(dp->d_name,"block_inserts")){
		    string fn = fname + "/" + dp->d_name;
		    process_nsrl_4k(b,fn);
		}
	    }
	    closedir(dirp);
	}
	else {
	    process_nsrl_4k(b,*i);
	}
    }
    
    if(opt_imagefile){
	process_imagefile(opt_imagefile,b,opt_sector_size);
	printf("%s:\n",opt_imagefile);
	printf("opt_blom_bits: %d\n",opt_bloom_bits);
	printf("opt_sector_size: %d\n",opt_sector_size);
    }

    print_report(b);
    nsrl_bloom_print_usage();
    if(outfile){
	cout << "writing " << outfile << "\n";
	if(b.write(outfile)){
	    err(1,"write failed");
	}
    }
    return 0;
}
#else
int main()
{
    fprintf(stderr,"disk_bloom currently requires AFFLIB\n");
    exit(1);
}
#endif
