#ifndef SHA1_T_H
#define SHA1_T_H

#include <assert.h>
#include <stdint.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <iostream>

using namespace std;

/* An efficient, object-oriented interface to OpenSSL.
 *
 * This instrumented sha1 keeps track of how many times it has been called and 
 * has a bypass for calculating SHA1 of all NULLs.
 * 
 * Uses the OpenSSL EVP class
 */

#ifdef HAVE_OPENSSL_HMAC_H
#include <openssl/hmac.h>
#else
#error Requires OpenSSL hmac.h
#endif

/*
 * Abstract base class for representing a hash code.
 */

class hashcode_t {
 public:
    const u_char *final() const {return 0;}	/* please override */
    size_t size() const {return 0;}
};

class sha1_t:hashcode_t {
 public:
    const u_char *final() const { return final_hash; }
    static const size_t SIZE = 20;
    size_t size() const{ return SIZE;}
    u_char final_hash[SIZE];
    sha1_t(){ }				/* somebody else must fill it in */
    /* Create a hash from a binary value. It can only be used for comparision */
    sha1_t(const u_char provided_hash[20]){
	memcpy(final_hash,provided_hash,sizeof(provided_hash));
    }
    /** Return the final value of the hash (finalize if necessary)
     */
    const char *hexdigest(char *hexbuf,size_t bufsize) const {
	const char *hexbuf_start = hexbuf;
	for(u_int i=0;i<SIZE && bufsize>3;i++){
	    snprintf(hexbuf,bufsize,"%02x",final_hash[i]);
	    hexbuf  += 2;
	    bufsize -= 2;
	    
	}
	return hexbuf_start;
    }
    string hexdigest() const {
	string ret;
	char buf[3];
	for(u_int i=0;i<SIZE;i++){
	    snprintf(buf,sizeof(buf),"%02x",final_hash[i]);
	    ret += buf;
	}
	return ret;
    }
    /* Create a sha1_t from a hex representation.
     */
    static const sha1_t *new_from_hex(const char *hex) {
	sha1_t *val = new sha1_t();
	if(nsrl_hex2bin(val->final_hash,sizeof(val->final_hash),hex)!=160){
	    cerr << "invalid sha1: " << hex << "\n";
	    exit(1);
	}
	return val;
    }

    /** Operator makes this class work with STL maps. Unfortunately, that means that they need to take
     * const, which means that they lie about the mutability change.
     */
    bool operator<(const sha1_t &s2) const 
    {
	return memcmp(this->final_hash,s2.final_hash, SIZE) < 0;
    }
    bool operator==(const sha1_t &s2) const
    {
	return memcmp(this->final_hash,s2.final_hash, SIZE) == 0;
    }
};


class hashgen_t { 			/* generates the hash */
 public:
    /* Static function to determine if something is zero */
    static bool iszero(const u_char *buf,size_t bufsize){
	for(u_int i=0;i<bufsize;i++){
	    if(buf[i]!=0) return false;
	}
	return true;
    }
 hashgen_t(const EVP_MD *md_):md(md_),initialized(false),finalized(false){
    }
    ~hashgen_t(){
	release();
    }
    const EVP_MD *md;
    EVP_MD_CTX mdctx;			// the context for computing the value
    bool initialized;				/* has the context been initialized? */
    bool finalized;
    size_t size(){				/* return the size of the hash */
	return EVP_MD_size(md);
    }
    void init(){
	if(initialized==false){
	    EVP_MD_CTX_init(&mdctx);
	    EVP_DigestInit_ex(&mdctx, md, NULL);
	    initialized = true;
	    finalized = false;
	}
    }
    void update(const u_char *buf,size_t bufsize){
	if(!initialized) init();
	if(finalized){
	    fprintf(stderr,"sha1_t::update called after finalized\n");
	    exit(1);
	}
	EVP_DigestUpdate(&mdctx,buf,bufsize);
    }
    void release(){			/* free allocated memory */
	if(initialized){
	    EVP_MD_CTX_cleanup(&mdctx);
	    initialized = false;
	}
    }


};

static uint64_t sha1_ctr=0;			// number of SHA1 operations performed
static u_char sha1_512nulls[20];		/* hash of a null block, as an optimization */
class sha1gen_t : hashgen_t {
 public:
 sha1gen_t():hashgen_t(EVP_sha1()){ }
    sha1_t final() {
	if(!finalized){
	    sha1_t val;
	    if(!initialized){
		fprintf(stderr,"sha1_t::final called before init!n");
		exit(1);
	    }
	    unsigned int len = sizeof(val.final_hash);
	    EVP_DigestFinal(&mdctx,val.final_hash,&len);
	    finalized = true;
	    return val;
	}
	cerr << "currently sha1gen_t does not cache the final value\n";
	exit(1);
    }

    /** Compute a sha1 from a buffer and return the hash */
    static sha1_t hash_buf(const u_char *buf,size_t bufsize){
	/* First time through find the SHA1 of 512 NULLs */
	if(sha1_ctr==0){
	    u_char b2[512];
	    EVP_MD_CTX ctx2;
	    unsigned int len = sizeof(sha1_512nulls);
	    memset(b2,0,sizeof(b2));
	    EVP_MD_CTX_init(&ctx2);
	    EVP_DigestInit_ex(&ctx2, EVP_sha1(), NULL);
	    EVP_DigestUpdate(&ctx2,b2,sizeof(b2));
	    EVP_DigestFinal(&ctx2,sha1_512nulls,&len);
	    sha1_ctr++;
	}

	/* If the input is 512 bytes long and all NULLs, use our SHA1 of 512 NULLs */
	if(bufsize==512 && iszero(buf,bufsize)){
	    return sha1_t(sha1_512nulls);
	}
	sha1gen_t g;
	g.update(buf,bufsize);
	return g.final();
    }
	
    /** Static method allocateor */
    static sha1_t hash_file(const char *fname){
	int fd = open(fname,O_RDONLY);
	if(fd<0) throw fname;
	struct stat st;
	if(fstat(fd,&st)<0){
	    close(fd);
	    throw fname;
	}
	const u_char *buf = (const u_char *)mmap(0,st.st_size,PROT_READ,MAP_FILE|MAP_SHARED,fd,0);
	if(buf==0){
	    close(fd);
	    throw fname;
	}
	sha1_t s = hash_buf(buf,st.st_size);
	munmap((void *)buf,st.st_size);
	close(fd);
	return s;
    }

};


#endif
